#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <ogcsys.h>
#include <fat.h>
#include <sdcard/wiisd_io.h>

#include "state.h"
#include "certs.h"
#include "fat.h"
#include "wiibasics.h"

/* Externs */
extern void unstub_start(u32);

/* Dol Header structure */
typedef struct {
	u32 textoff[7];
	u32 dataoff[11];
	u8 *textmem[7];
	u8 *datamem[11];
	u32 textsize[7];
	u32 datasize[11];
	u8 *bssmem;
	u32 bsssize;
	u32 entry;
	u32 unused[7];
} dolhdr;

/* Ticket structure */
typedef struct {
	sig_rsa2048 signature;
	tik tik_data;
} ticket;

/* Constants */
#define SYSMENU_ADDR		0x8132FF80
#define SYSMENU_LEN		0x380000
#define SYSMENU_TITLEID		0x100000002ULL

// Preloader-style hacks file
#define HACKS_FILE "banana_hacks.txt"


char* substr(char* oldString,int start,int length){
	char *newString=malloc(length*(sizeof(char))+1);
	int i;
	for(i=0;i<length;i++){
		newString[i]=oldString[start+i];
	}
	newString[length]='\0';

	return newString;
}


bool Read_And_Apply_Patches(u32 smversion){

	char *archivoLeido=NULL;
	__io_wiisd.startup();
	fatMountSimple("sd", &__io_wiisd);
	if(!fatInitDefault()){
		printf("- ERROR: I can't initialize SD. No card inserted?\n");
		return false;
	}
	char filepath[128];
	sprintf(filepath, "fat0:/" HACKS_FILE);

	s32 ret = Fat_ReadFile(filepath, (void *)&archivoLeido);

	if (ret > 0) {
		u32 start=0;
		u32 length=0;
		u32 maxLength=strlen(archivoLeido);
		u32 codesStack[32];
		u8 stackIn=0;
		u8 stackOut=0;
		printf("- sd:/%s found.",HACKS_FILE);
		bool enabled=false;
		bool endOfFile=false;
		while(!endOfFile){
			length++;
			if(start+length==maxLength){
				endOfFile=true;
			}
			if((archivoLeido[start+length]==13 && archivoLeido[start+length+1]==10) || endOfFile){
				char* cortado=substr(archivoLeido,start,length);
				if(cortado[0]=='[' && cortado[length-1]==']'){
					if(enabled){
						printf("OK!");
					}
					printf("\n  - Found %s patch... ",cortado);
					stackIn=0;
					stackOut=0;
					enabled=true;
				}else if(length>8){
					char* tag=substr(cortado,0,7);
					if(strcmp(tag,"version")==0){
						int codeversion=atoi(substr(cortado,8,length-8));
						if(smversion!=codeversion){
							enabled=false;
							printf(" Invalid version.");
						}
					}else if(strcmp(tag,"offset=")==0){
						int i;
						for(i=7;i<length;i+=11){
							u32 hex=0;
							sscanf(substr(cortado,i,10), "%X", &hex);
							codesStack[stackIn++]=hex;
						}
					}else if(strcmp(tag,"value=0")==0 && enabled==true){
						int i;
						for(i=6;i<length;i+=11){
							u32 offset=codesStack[stackOut++];
							u32 value=0;
							sscanf(substr(cortado,i,10), "%X", &value);
							*(u32 *)offset=value;
						}
					}
					
				}
				
				start=start+length+2;
				length=0;
			}
		}
		if(enabled){
			printf("OK!");
		}

		free(archivoLeido);
		fatUnmount("sd");
		__io_wiisd.shutdown();
		return true;
	}else{
		printf("- ERROR: I can't find sd:/%s patches file.\n",HACKS_FILE);
		return false;
	}
//exit(0);
}

s32 Sysmenu_Load(u32 *entry, u16 index)
{
	static dolhdr dol ATTRIBUTE_ALIGN(32);

	u32 cnt;
	s32 fd, ret;

	/* Open content */
	fd = ES_OpenContent(index);
	if (fd < 0)
		return fd;

	/* Read content */
	ret = ES_ReadContent(fd, (u8 *)&dol, sizeof(dolhdr));
	if (ret < 0)
		return ret;

	memset(dol.bssmem, 0, dol.bsssize);

	/* Read data */
	for (cnt = 0; cnt < 7; cnt++)
		if (dol.textoff[cnt] >= sizeof(dolhdr)) {
			ret = ES_SeekContent(fd, dol.textoff[cnt], 0);
			if (ret != dol.textoff[cnt])
				return -1;

			ret = ES_ReadContent(fd, dol.textmem[cnt], dol.textsize[cnt]);
			if (ret != dol.textsize[cnt])
				return -1;
		}

	for (cnt = 0; cnt < 11; cnt++)
		if (dol.dataoff[cnt] >= sizeof(dolhdr)) {
			ret = ES_SeekContent(fd, dol.dataoff[cnt], 0);
			if (ret != dol.dataoff[cnt])
				return -1;

			ret = ES_ReadContent(fd, dol.datamem[cnt], dol.datasize[cnt]);
			if (ret != dol.datasize[cnt])
				return -1;
		}

	/* Close content */
	ES_CloseContent(fd);

	/* Set entry point */
	*entry = dol.entry;

	return 0;
}

s32 Sysmenu_Identify(void)
{
	static ticket s_tik ATTRIBUTE_ALIGN(32);
	signed_blob *p_certs = NULL, *p_tik = NULL, *p_tmd = NULL;

	u32 certs_len, tik_len, tmd_len;
	s32 ret;

	/* Retrieve certificates */
	ret = Sys_GetCerts(&p_certs, &certs_len);
	if (ret < 0)
		return ret;

	/* Set ticket length */
	tik_len = sizeof(s_tik);

	memset(&s_tik, 0, tik_len);

	/* Generate ticket */
	strcpy(s_tik.tik_data.issuer, "Root-CA00000001-XS00000003");
	memset(s_tik.tik_data.cidx_mask, 0xFF, 32);

	s_tik.signature.type = ES_SIG_RSA2048;

	/* Set pointer */
	p_tik = (signed_blob *)&s_tik;

	/* Retrieve TMD length */
	ret = ES_GetStoredTMDSize(SYSMENU_TITLEID, &tmd_len);
	if (ret < 0)
		goto out;

	/* Allocate memory */
	p_tmd = (signed_blob *)memalign(32, tmd_len);
	if (!p_tmd) {
		ret = -1;
		goto out;
	}

	/* Retrieve TMD */
	ret = ES_GetStoredTMD(SYSMENU_TITLEID, p_tmd, tmd_len);
	if (ret < 0)
		goto out;

	/* Identify as system menu */
	ret = ES_Identify(p_certs, certs_len, p_tmd, tmd_len, p_tik, tik_len, NULL);

out:
	/* Free memory */
	if (p_tmd)
		free(p_tmd);

	return ret;
}

s32 Sysmenu_Bootindex(void)
{
	signed_blob *p_tmd = NULL;

	u32 len;
	s32 ret;

	/* Retrieve TMD length */
	ret = ES_GetStoredTMDSize(SYSMENU_TITLEID, &len);
	if (ret < 0)
		goto out;

	/* Allocate memory */
	p_tmd = (signed_blob *)memalign(32, len);
	if (!p_tmd)
		return -1;

	/* Retrieve TMD */
	ret = ES_GetStoredTMD(SYSMENU_TITLEID, p_tmd, len);
	if (ret < 0)
		goto out;

	/* Get boot index */
	ret = ((tmd *)SIGNATURE_PAYLOAD(p_tmd))->boot_index;

out:
	/* Free memory */
	if (p_tmd)
		free(p_tmd);

	return ret;
}

u32 get_sysmenu_version(){
	s32 ret;
	u32 tmd_size;

	ret=ES_GetStoredTMDSize(TITLE_ID(1,2), &tmd_size);
	static u8 tmd_buf[MAX_SIGNED_TMD_SIZE] ATTRIBUTE_ALIGN(32);
	signed_blob *s_tmd=(signed_blob *)tmd_buf;
	ret=ES_GetStoredTMD(TITLE_ID(1,2), s_tmd, tmd_size);

	tmd *t=SIGNATURE_PAYLOAD(s_tmd);
	u32 tidh=t->title_id >> 32;
	u32 tidl=t->title_id & 0xFFFFFFFF;

	if(tidh==1 && tidl==2 && t->title_version){
		printf("- Found System Menu (v%u - IOS%u).\n",(u32)t->title_version,(u32)t->sys_version);
		return (u32)t->title_version;
	}else{
		return 0;
	}
}

s32 Sysmenu_Launch(void)
{
	u32 entry, index;
	s32 ret;
	u32 smversion;

	/* Read current state flags */
	ret = State_Initialize();
	if (ret < 0) {
		printf("- ERROR: Could not read current state flags! (ret = %d)\n", ret);
		return ret;
	}

	/* Identify as system menu */
	ret = Sysmenu_Identify();
	if (ret < 0) {
		printf("- ERROR: Could not identify as system menu! (ret = %d)\n", ret);
		return ret;
	}

	// Get system menu version
	smversion=get_sysmenu_version();
	if(smversion==0){
		printf("- ERROR: I can't get System Menu version.\n");
		return -1;
	}

	printf("- Loading System Menu into memory...");
	fflush(stdout);

	/* Get system menu boot index */
	ret = Sysmenu_Bootindex();
	if (ret < 0) {
		printf("- ERROR: Could not retrieve boot index! (ret = %d)\n", ret);
		return ret;
	}

	index = ret;

	/* Load system menu */
	ret = Sysmenu_Load(&entry, index);
	if (ret < 0) {
		printf("- ERROR: Could not load system menu executable! (ret = %d)\n", ret);
		return ret;
	}
	printf(" OK!\n");

	/* Patch system menu */
	bool last_step=Read_And_Apply_Patches(smversion);
	if(!last_step)
		return -2;

	/* Set state flags */
	ret = State_ReturnToMenu();
	if (ret < 0) {
		printf("- ERROR: Could not set state flags! (ret = %d)\n", ret);
		return ret;
	}

	SYS_ResetSystem(SYS_SHUTDOWN, 0, 0);
	DCFlushRange((u8 *)SYSMENU_ADDR, SYSMENU_LEN);

	/* Jump to entry point */
	unstub_start(entry);

	printf(" returned!\n");

	return 0;
}
